package pomocov;


import beast.evolution.likelihood.LikelihoodCore;
import pomocov.utils.IndicesEtc;

import static pomocov.utils.IndicesEtc.*;

/**
 * standard likelihood core, uses no caching *
 */



// GRJ  TODO It is not clear to me if extending LikelihoodCore is helpful.

public class BeerLikelihoodCorePoMo extends LikelihoodCore {
    protected int nrOfStates;
    protected int nrOfNodes;
    protected int nrOfPatterns;
    protected int partialsSize;
    protected int matrixSize;
    protected int nrOfMatrices;

    protected boolean integrateCategories;

    protected double[][][] partials;

    protected int[][] states;

    protected double[][][][] matrices;

    protected int[] currentMatrixIndex;
    protected int[] storedMatrixIndex;
    protected int[] currentPartialsIndex;
    protected int[] storedPartialsIndex;

    protected boolean useScaling = false;

    protected double[][][] scalingFactors;

    private double scalingThreshold = 1.0E-100;
    double SCALE = 2;


    public BeerLikelihoodCorePoMo(int nrOfStates) {
        this.nrOfStates = nrOfStates;
    } // c'tor


    /**
     * Integrates partials across categories.
     *
     * @param inPartials  the array of partials to be integrated
     * @param proportions the proportions of sites in each category
     * @param outPartials an array into which the partials will go
     */
    @Override
	protected void calculateIntegratePartials(double[] inPartials, double[] proportions, double[] outPartials) {

        int u = 0;
        int v = 0;
        for (int k = 0; k < nrOfPatterns; k++) {

            for (int i = 0; i < nrOfStates; i++) {

                outPartials[u] = inPartials[v] * proportions[0];
                u++;
                v++;
            }
        }


        for (int l = 1; l < nrOfMatrices; l++) {
            u = 0;

            for (int k = 0; k < nrOfPatterns; k++) {

                for (int i = 0; i < nrOfStates; i++) {

                    outPartials[u] += inPartials[v] * proportions[l];
                    u++;
                    v++;
                }
            }
        }
    }

    /**
     * Calculates pattern log likelihoods at a node.
     *
     * @param partials          the partials used to calculate the likelihoods
     * @param frequencies       an array of state frequencies
     * @param outLogLikelihoods an array into which the likelihoods will go
     */
    @Override
	public void calculateLogLikelihoods(double[] partials, double[] frequencies, double[] outLogLikelihoods) {
        int v = 0;
        for (int k = 0; k < nrOfPatterns; k++) {

            double sum = 0.0;
            for (int i = 0; i < nrOfStates; i++) {

                sum += frequencies[i] * partials[v];
                v++;
            }
            outLogLikelihoods[k] = Math.log(sum) + getLogScalingFactor(k);
        }
    }


    /**
     * initializes partial likelihood arrays.
     *
     * @param nodeCount           the number of nodes in the tree
     * @param patternCount        the number of patterns
     * @param matrixCount         the number of matrices (i.e., number of categories)
     * @param integrateCategories whether sites are being integrated over all matrices
     */
    @Override
	public void initialize(int nodeCount, int patternCount, int matrixCount, boolean integrateCategories, boolean useAmbiguities) {

        assert matrixCount == 1; // GRJ checking.
        this.nrOfNodes = nodeCount;
        this.nrOfPatterns = patternCount;
        this.nrOfMatrices = matrixCount;

        this.integrateCategories = integrateCategories;

        if (integrateCategories) {
            partialsSize = patternCount * nrOfStates * matrixCount;
        } else {
            partialsSize = patternCount * nrOfStates;
        }

        partials = new double[2][nodeCount][];

        currentMatrixIndex = new int[nodeCount];
        storedMatrixIndex = new int[nodeCount];

        currentPartialsIndex = new int[nodeCount];
        storedPartialsIndex = new int[nodeCount];

        states = new int[nodeCount][];

        for (int i = 0; i < nodeCount; i++) {
            partials[0][i] = null;
            partials[1][i] = null;

            states[i] = null;
        }

        matrixSize = nrOfStates * nrOfStates;

        matrices = new double[2][nodeCount][][];
    }

    /**
     * cleans up and deallocates arrays.
     */
    @Override
	public void finalize() throws java.lang.Throwable {
        nrOfNodes = 0;
        nrOfPatterns = 0;
        nrOfMatrices = 0;

        partials = null;
        currentPartialsIndex = null;
        storedPartialsIndex = null;
        states = null;
        matrices = null;
        currentMatrixIndex = null;
        storedMatrixIndex = null;

        scalingFactors = null;
    }

    @Override
    public void setUseScaling(double scale) {
        useScaling = (scale != 1.0);

        if (useScaling) {
            scalingFactors = new double[2][nrOfNodes][nrOfPatterns];
        }
    }

    /**
     * Allocates partials for a node
     */
    @Override
	public void createNodePartials(int nodeIndex) {

        this.partials[0][nodeIndex] = new double[partialsSize];
        this.partials[1][nodeIndex] = new double[partialsSize];
    }

    /**
     * Sets partials for a node
     */
    @Override
	public void setNodePartials(int nodeIndex, double[] partials) {

        if (this.partials[0][nodeIndex] == null) {
            createNodePartials(nodeIndex);
        }
        if (partials.length < partialsSize) {
            int k = 0;
            for (int i = 0; i < nrOfMatrices; i++) {
                System.arraycopy(partials, 0, this.partials[0][nodeIndex], k, partials.length);
                k += partials.length;
            }
        } else {
            System.arraycopy(partials, 0, this.partials[0][nodeIndex], 0, partials.length);
        }
    }

    @Override
    public void getNodePartials(int nodeIndex, double[] partialsOut) {
        System.arraycopy(partials[currentPartialsIndex[nodeIndex]][nodeIndex], 0, partialsOut, 0, partialsOut.length);
    }

    /**
     * Allocates states for a node
     */
    public void createNodeStates(int nodeIndex) {

        this.states[nodeIndex] = new int[nrOfPatterns];
    }

    /**
     * Sets states for a node
     */
    @Override
	public void setNodeStates(int nodeIndex, int[] states) {

        if (this.states[nodeIndex] == null) {
            createNodeStates(nodeIndex);
        }
        System.arraycopy(states, 0, this.states[nodeIndex], 0, nrOfPatterns);
    }

    /**
     * Gets states for a node
     */
    @Override
	public void getNodeStates(int nodeIndex, int[] states) {
        System.arraycopy(this.states[nodeIndex], 0, states, 0, nrOfPatterns);
    }

    @Override
    public void setNodeMatrixForUpdate(int nodeIndex) {
        currentMatrixIndex[nodeIndex] = 1 - currentMatrixIndex[nodeIndex];

    }


    //NICOLA: actually store a matrix for each interval of interest within a node (the bottleneck is the same for all nodes,
    // so you can just use the one from TreeLikelihoodPoMo).
    /**
     * Sets probability matrix for a node
     */
    @Override
	public void setNodeMatrix(int nodeIndex, int matrixIndex, double[] matrix) {
    	System.out.println("Setting node Matrix1 ");
        System.arraycopy(matrix, 0, matrices[currentMatrixIndex[nodeIndex]][nodeIndex],
                matrixIndex * matrixSize, matrixSize);
    }


    /**
     * Gets probability matrix for a node
     */
    @Override
	public void getNodeMatrix(int nodeIndex, int matrixIndex, double[] matrix) {
        System.arraycopy(matrices[currentMatrixIndex[nodeIndex]][nodeIndex],
                matrixIndex * matrixSize, matrix, 0, matrixSize);
    }


    @Override
    public void setNodePartialsForUpdate(int nodeIndex) {
        currentPartialsIndex[nodeIndex] = 1 - currentPartialsIndex[nodeIndex];
    }


    @Override
	public void calculatePartials(int nodeIndex1, int nodeIndex2, int nodeIndex3) {
        // GRJ. BeerLikelihoodCorePoMo extends LikelihoodCore so must implement this,
        // but it is not used.
        assert false;
    }





    @Override
    public void integratePartials(int nodeIndex, double[] proportions, double[] outPartials) {
        calculateIntegratePartials(partials[currentPartialsIndex[nodeIndex]][nodeIndex], proportions, outPartials);
    }


    /**
     * This function returns the scaling factor for that pattern by summing over
     * the log scalings used at each node. If scaling is off then this just returns
     * a 0.
     *
     * @return the log scaling factor
     */
    @Override
    public double getLogScalingFactor(int patternIndex_) {
        double logScalingFactor = 0.0;
        if (useScaling) {
            for (int i = 0; i < nrOfNodes; i++) {
                logScalingFactor += scalingFactors[currentPartialsIndex[i]][i][patternIndex_];
            }
        }
        return logScalingFactor;
    }


    /**
     * Store current state
     */
    @Override
    public void restore() {
        // Rather than copying the stored stuff back, just swap the pointers...
        int[] tmp1 = currentMatrixIndex;
        currentMatrixIndex = storedMatrixIndex;
        storedMatrixIndex = tmp1;

        int[] tmp2 = currentPartialsIndex;
        currentPartialsIndex = storedPartialsIndex;
        storedPartialsIndex = tmp2;
    }

    @Override
    public void unstore() {
        System.arraycopy(storedMatrixIndex, 0, currentMatrixIndex, 0, nrOfNodes);
        System.arraycopy(storedPartialsIndex, 0, currentPartialsIndex, 0, nrOfNodes);
    }

    /**
     * Restore the stored state
     */
    @Override
    public void store() {
        System.arraycopy(currentMatrixIndex, 0, storedMatrixIndex, 0, nrOfNodes);
        System.arraycopy(currentPartialsIndex, 0, storedPartialsIndex, 0, nrOfNodes);
    }



    /* ****************************************************************************************************************/
    /* *********************************     main calculation per pattern    ******************************************/
    /* ****************************************************************************************************************/

    /*** Update partial likelihoods at a nodePoMo when meeting a child or sample. */
    /* GRJ. This is orig code. Used for intervals within a node
                    partials1
                        |
                        |  matrix = matrices1[0]
                        |
                        * partials2
                    partials3
     */
    protected void calculatePartialsFixedPruning(double[] partials1, double[] matrices1, double[] partials2, double[] partials3) {
        double sum1;
        int u = 0;
        int v = 0;
        for (int l = 0; l < nrOfMatrices; l++) {
            for (int k = 0; k < nrOfPatterns; k++) {
                int w = l * matrixSize;
                for (int i = 0; i < nrOfStates; i++) {
                    sum1 = 0.0;
                    for (int j = 0; j < nrOfStates; j++) {
                        sum1 += matrices1[w] * partials1[v + j];
                        w++;
                    }
                    partials3[u] = sum1 * partials2[u];
                    u++;
                }
                v += nrOfStates;
            }
        }
    }



    /*** Update partial likelihoods at a nodePoMo only considering an interval (one branch only). */
     /* GRJ. This is orig code. Used for last interval to node, and bottleneck
                    partials1
                        |
                        |  matrix = matrices1[0]
                        |
                    partials3
      */
    protected void calculatePartialsPruning(double[] partials1, double[] matrices1, double[] partials3) {
        double sum1;
        int u = 0;
        int v = 0;
        for (int l = 0; l < nrOfMatrices; l++) {
            for (int k = 0; k < nrOfPatterns; k++) {
                int w = l * matrixSize;
                for (int i = 0; i < nrOfStates; i++) {
                    sum1 = 0.0;
                    for (int j = 0; j < nrOfStates; j++) {
                        sum1 += matrices1[w] * partials1[v + j];
                        w++;
                    }
                    partials3[u] = sum1;
                    u++;
                }
                v += nrOfStates;
            }
        }
    }


    /* GRJ. Used for intervals within a node. Linear approx.
                partials1
                    |
                    |   mutPs[]
                    |
                    * partials2
                partials3
    */
    protected void calcPartialsToChOrSmp_Linear(double[] partials1, double[] mutPs, double[] partials2, double[] partials3, int vPop) {
        calcPartialsOneBranch_Linear(partials1, mutPs, partials3, vPop);
        for (int i = 0; i < nrOfStates; i++) {
            partials3[i] *= partials2[i];
        }
    }


    /* Used for last interval to node. Linear approx.
                   partials1
                       |
                       |  mutPs[]
                       |
                   partials3
    */
    protected void calcPartialsOneBranch_Linear(double[] partials1, double[] mutPs, double[] partials3, int vPop) {
        // Example. For a Jukes-Cantor, the mutPs[] will be all
        //      p = mu * d * (1/3)
        // where mu is mutation rate per site per day (~= 2.5e-6) and d is number of days (~=5).
        // Then, for example,
        //     P(observe all As | all As at node)  =  1.0 - p - p - p  =  1.0 - (mu * d)
        assert mutPs.length == 12;
        for (int ii = 0; ii < 12; ii++) {
            assert mutPs[ii] >= 0.0;
            assert mutPs[ii] < 1.0 / (3.0);
        }
        int sA1Cn = IndicesEtc.stateA1Cn(vPop);
        int sAnC1 = IndicesEtc.stateAnC1(vPop);
        int sA1Gn = IndicesEtc.stateA1Gn(vPop);
        int sAnG1 = IndicesEtc.stateAnG1(vPop);
        int sA1Tn = IndicesEtc.stateA1Tn(vPop);
        int sAnT1 = IndicesEtc.stateAnT1(vPop);
        int sC1Gn = IndicesEtc.stateC1Gn(vPop);
        int sCnG1 = IndicesEtc.stateCnG1(vPop);
        int sC1Tn = IndicesEtc.stateC1Tn(vPop);
        int sCnT1 = IndicesEtc.stateCnT1(vPop);
        int sG1Tn = IndicesEtc.stateG1Tn(vPop);
        int sGnT1 = IndicesEtc.stateGnT1(vPop);

        // The conditional probabilities for the polymorphic states is unchanged by
        // mutation given this model and approximation. The matrix of transition
        // probabilities for mutation-only is like
        //            A B
        //            0 I
        // where A is 4 by 4, 0 is all zeros, I is identity, and
        // B is 4 by 6*(vPop-1) with 4*3 nonzero entries
        System.arraycopy(partials1, 0, partials3, 0, partials1.length);
        for (int k = 0; k < nrOfPatterns; k++) {
            int u = k * nrOfStates;
            partials3[u + sAAA] *= (1.0 - mutPs[rAC] - mutPs[rAG] - mutPs[rAT]);
            partials3[u + sAAA] += mutPs[rAC] * partials1[u + sAnC1];
            partials3[u + sAAA] += mutPs[rAG] * partials1[u + sAnG1];
            partials3[u + sAAA] += mutPs[rAT] * partials1[u + sAnT1];

            partials3[u + sCCC] *= (1.0 - mutPs[rCA] - mutPs[rCG] - mutPs[rCT]);
            partials3[u + sCCC] += mutPs[rCA] * partials1[u + sA1Cn];
            partials3[u + sCCC] += mutPs[rCG] * partials1[u + sCnG1];
            partials3[u + sCCC] += mutPs[rCT] * partials1[u + sCnT1];

            partials3[u + sGGG] *=  (1.0 - mutPs[rGA] - mutPs[rGC] - mutPs[rGT]);
            partials3[u + sGGG] += mutPs[rGA] * partials1[u + sA1Gn];
            partials3[u + sGGG] += mutPs[rGC] * partials1[u + sC1Gn];
            partials3[u + sGGG] += mutPs[rGT] * partials1[u + sGnT1];

            partials3[u + sTTT] *=  (1.0 - mutPs[rTA] - mutPs[rTC] - mutPs[rTG]);
            partials3[u + sTTT] += mutPs[rTA] * partials1[u + sA1Tn];
            partials3[u + sTTT] += mutPs[rTC] * partials1[u + sC1Tn];
            partials3[u + sTTT] += mutPs[rTG] * partials1[u + sG1Tn];
            // GRJ TODO no need to arraycopy, could update one partials[]
        }
    }



    /*** GRJ Like calculatePartialsPruning()
     * Bottleneck using Drift-2-nucleotides matrix drift2nucsMatrix[]
     * drift2nucsMatrix[] is (nrOfStates+1)^2 matrix used for each of A-C, A-G, , , , G-T*/
    // GRJ TODO use IndicesEtc to simplify
    protected void calcPartialsBottle_D2N(double[] partials1, double[] drift2nucsMatrix, double[] partials3, int vPop) {
        assert nrOfMatrices == 1;
        assert nrOfStates == IndicesEtc.nPoMoStates(vPop);
        assert drift2nucsMatrix.length == (vPop + 1) * (vPop + 1);
        // A to C.  (0, 0*vPop+4,0*vPop+5,...1*vPop+2, 1)
        // A to G.  (0, 1*vPop+3,1*vPop+4,...2*vPop+1, 2)
        // A to T.  (0, 2*vPop+2,1*vPop+3,...3*vPop+0, 3)
        // C to G.  (1, 3*vPop+1,3*vPop+1,...4*vPop-1, 2)
        // C to T.  (1, 4*vPop+0,4*vPop+1,...5*vPop-2, 3)
        // G to T.  (2, 5*vPop-2,5*vPop-1,...6*vPop-3, 3)
        int[] vtxs0 = {0,0,0,1,1,2};
        int[] vtxs1 = {1,2,3,2,3,3};
        int vPopm1 = vPop - 1;
        int vPopp1 = vPop + 1;
        for (int k = 0; k < nrOfPatterns; k++) {
            int u = k * nrOfStates;
            for (int i = 0; i < 4; i++) {
                partials3[u + i] = partials1[u + i];
            }
            for (int d = 0; d < 6; d++) {
                int vtx0 = vtxs0[d];
                int vtx1 = vtxs1[d];
                int s = u + 4 + d*vPopm1;
                for (int i = 0; i < vPop-1; i++) {
                    double sum = 0.0;
                    int x = (i+1) * vPopp1;
                    sum += drift2nucsMatrix[x + vPop] * partials1[u + vtx0];
                    sum += drift2nucsMatrix[x] * partials1[u + vtx1];
                    for (int j = 0; j < vPop-1; j++) {
                        sum += drift2nucsMatrix[x + 1 + j] * partials1[s + j];
                    }
                    partials3[s + i] = sum;
                }
            }
        }
    }


    // 2020-06-24. GRJ I'm unconvinced this is an improvement. Going back to Moran drift calculation for now
    // To reinstate, call from PoMoCoVTreeLikelihood.calcBottleneckPartials() instead of
    // calcPartialsBottle_D2N_V3() and use Bottleneck.bottleneckForVPop3().
    // TODO remove if not used
    void calcPartialsBottle_Binomial_V3(double[] partials1, double[] bprobs, double[] partials3) {
        assert bprobs.length == 4;
        double p30 = bprobs[3];
        double p21 = bprobs[2];
        double p12 = bprobs[1];
        double p03 = bprobs[0];

        for (int k = 0; k < nrOfPatterns; k++) {
            int u = k * nrOfStates;
            partials3[u + sAAA] = partials1[u + sAAA];
            partials3[u + sCCC] = partials1[u + sCCC];
            partials3[u + sGGG] = partials1[u + sGGG];
            partials3[u + sTTT] = partials1[u + sTTT];

            partials3[u + sACC] = p30 * partials1[u + sAAA] + p21 * partials1[u + sAAC] + p12 * partials1[u + sACC] + p03 * partials1[u + sCCC];
            partials3[u + sAAC] = p03 * partials1[u + sAAA] + p12 * partials1[u + sAAC] + p21 * partials1[u + sACC] + p30 * partials1[u + sCCC];

            partials3[u + sAGG] = p30 * partials1[u + sAAA] + p21 * partials1[u + sAAG] + p12 * partials1[u + sAGG] + p03 * partials1[u + sGGG];
            partials3[u + sAAG] = p03 * partials1[u + sAAA] + p12 * partials1[u + sAAG] + p21 * partials1[u + sAGG] + p30 * partials1[u + sGGG];

            partials3[u + sATT] = p30 * partials1[u + sAAA] + p21 * partials1[u + sAAT] + p12 * partials1[u + sATT] + p03 * partials1[u + sTTT];
            partials3[u + sAAT] = p03 * partials1[u + sAAA] + p12 * partials1[u + sAAT] + p21 * partials1[u + sATT] + p30 * partials1[u + sTTT];

            partials3[u + sCGG] = p30 * partials1[u + sCCC] + p21 * partials1[u + sCCG] + p12 * partials1[u + sCGG] + p03 * partials1[u + sGGG];
            partials3[u + sCCG] = p03 * partials1[u + sCCC] + p12 * partials1[u + sCCG] + p21 * partials1[u + sCGG] + p30 * partials1[u + sGGG];

            partials3[u + sCTT] = p30 * partials1[u + sCCC] + p21 * partials1[u + sCCT] + p12 * partials1[u + sCTT] + p03 * partials1[u + sTTT];
            partials3[u + sCCT] = p03 * partials1[u + sCCC] + p12 * partials1[u + sCCT] + p21 * partials1[u + sCTT] + p30 * partials1[u + sTTT];

            partials3[u + sGTT] = p30 * partials1[u + sGGG] + p21 * partials1[u + sGGT] + p12 * partials1[u + sGTT] + p03 * partials1[u + sTTT];
            partials3[u + sGGT] = p03 * partials1[u + sGGG] + p12 * partials1[u + sGGT] + p21 * partials1[u + sGTT] + p30 * partials1[u + sTTT];
        }
    }



    /* Bottleneck using 4 values from Drift-2-nucleotides matrix, specialised for V=3 */
    protected void calcPartialsBottle_D2N_V3(double[] partials1, double[] drift2NucsMatrix, double[] partials3) {
        assert drift2NucsMatrix.length == 4*4;
        double p30 = drift2NucsMatrix[7];
        double p21 = drift2NucsMatrix[6];
        double p12 = drift2NucsMatrix[5];
        double p03 = drift2NucsMatrix[4];
        // GRJ TODO I don't understand ordering of e-vectors, shouldn't rely on it.

        for (int k = 0; k < nrOfPatterns; k++) {
            int u = k * nrOfStates;
            partials3[u + sAAA] = partials1[u + sAAA];
            partials3[u + sCCC] = partials1[u + sCCC];
            partials3[u + sGGG] = partials1[u + sGGG];
            partials3[u + sTTT] = partials1[u + sTTT];

            partials3[u + sACC] = p30 * partials1[u + sAAA] + p21 * partials1[u + sAAC] + p12 * partials1[u + sACC] + p03 * partials1[u + sCCC];
            partials3[u + sAAC] = p03 * partials1[u + sAAA] + p12 * partials1[u + sAAC] + p21 * partials1[u + sACC] + p30 * partials1[u + sCCC];

            partials3[u + sAGG] = p30 * partials1[u + sAAA] + p21 * partials1[u + sAAG] + p12 * partials1[u + sAGG] + p03 * partials1[u + sGGG];
            partials3[u + sAAG] = p03 * partials1[u + sAAA] + p12 * partials1[u + sAAG] + p21 * partials1[u + sAGG] + p30 * partials1[u + sGGG];

            partials3[u + sATT] = p30 * partials1[u + sAAA] + p21 * partials1[u + sAAT] + p12 * partials1[u + sATT] + p03 * partials1[u + sTTT];
            partials3[u + sAAT] = p03 * partials1[u + sAAA] + p12 * partials1[u + sAAT] + p21 * partials1[u + sATT] + p30 * partials1[u + sTTT];

            partials3[u + sCGG] = p30 * partials1[u + sCCC] + p21 * partials1[u + sCCG] + p12 * partials1[u + sCGG] + p03 * partials1[u + sGGG];
            partials3[u + sCCG] = p03 * partials1[u + sCCC] + p12 * partials1[u + sCCG] + p21 * partials1[u + sCGG] + p30 * partials1[u + sGGG];

            partials3[u + sCTT] = p30 * partials1[u + sCCC] + p21 * partials1[u + sCCT] + p12 * partials1[u + sCTT] + p03 * partials1[u + sTTT];
            partials3[u + sCCT] = p03 * partials1[u + sCCC] + p12 * partials1[u + sCCT] + p21 * partials1[u + sCTT] + p30 * partials1[u + sTTT];

            partials3[u + sGTT] = p30 * partials1[u + sGGG] + p21 * partials1[u + sGGT] + p12 * partials1[u + sGTT] + p03 * partials1[u + sTTT];
            partials3[u + sGGT] = p03 * partials1[u + sGGG] + p12 * partials1[u + sGGT] + p21 * partials1[u + sGTT] + p30 * partials1[u + sTTT];
        }
    }







    }




    /* Might want to revive this idea.See notes 16 Jun 2020
      Transition probabilities for bottleneck of size 3 where states account for regions in tetrahedron.
           AAA    CCC    GGG    TTT     ACC     AAC     AGG     AAG     ATT     AAT     CGG     CCG     CTT     CCT     GTT     GGT
    AAA 0.9985 0.0000 0.0000 0.0000 0.00000 0.00050 0.00000 0.00050 0.00000 0.00050 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000
    CCC 0.0000 0.9985 0.0000 0.0000 0.00050 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000 0.00050 0.00000 0.00050 0.00000 0.00000
    GGG 0.0000 0.0000 0.9985 0.0000 0.00000 0.00000 0.00050 0.00000 0.00000 0.00000 0.00050 0.00000 0.00000 0.00000 0.00000 0.00050
    TTT 0.0000 0.0000 0.0000 0.9985 0.00000 0.00000 0.00000 0.00000 0.00050 0.00000 0.00000 0.00000 0.00050 0.00000 0.00050 0.00000
    ACC 0.0405 0.3380 0.0000 0.0000 0.41400 0.20700 0.00000 0.00000 0.00000 0.00000 0.00000 0.00025 0.00000 0.00025 0.00000 0.00000
    AAC 0.3380 0.0405 0.0000 0.0000 0.20700 0.41400 0.00000 0.00025 0.00000 0.00025 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000
    AGG 0.0405 0.0000 0.3380 0.0000 0.00000 0.00000 0.41400 0.20700 0.00000 0.00000 0.00025 0.00000 0.00000 0.00000 0.00000 0.00025
    AAG 0.3380 0.0000 0.0405 0.0000 0.00000 0.00025 0.20700 0.41400 0.00000 0.00025 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000
    ATT 0.0405 0.0000 0.0000 0.3380 0.00000 0.00000 0.00000 0.00000 0.41400 0.20700 0.00000 0.00000 0.00025 0.00000 0.00025 0.00000
    AAT 0.3380 0.0000 0.0000 0.0405 0.00000 0.00025 0.00000 0.00025 0.20700 0.41400 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000
    CGG 0.0000 0.0405 0.3380 0.0000 0.00000 0.00000 0.00025 0.00000 0.00000 0.00000 0.41400 0.20700 0.00000 0.00000 0.00000 0.00025
    CCG 0.0000 0.3380 0.0405 0.0000 0.00025 0.00000 0.00000 0.00000 0.00000 0.00000 0.20700 0.41400 0.00000 0.00025 0.00000 0.00000
    CTT 0.0000 0.0405 0.0000 0.3380 0.00000 0.00000 0.00000 0.00000 0.00025 0.00000 0.00000 0.00000 0.41400 0.20700 0.00025 0.00000
    CCT 0.0000 0.3380 0.0000 0.0405 0.00025 0.00000 0.00000 0.00000 0.00000 0.00000 0.00000 0.00025 0.20700 0.41400 0.00000 0.00000
    GTT 0.0000 0.0000 0.0405 0.3380 0.00000 0.00000 0.00000 0.00000 0.00025 0.00000 0.00000 0.00000 0.00025 0.00000 0.41400 0.20700
    GGT 0.0000 0.0000 0.3380 0.0405 0.00000 0.00000 0.00025 0.00000 0.00000 0.00000 0.00025 0.00000 0.00000 0.00000 0.20700 0.41400


    // Bottleneck of fixed size 3, for V=3, where states account for regions in tetrahedron.
    //  There are 4*4 + 12*6 = 88 nonzero entries

    protected void calcPartialsBottle_V3_88nonzero(double[] ps, double[] partials3) {
        double pXXX_XXX = 0.9985;
        double pXXX_XXY = 0.00050;

        double pXYY_XXX = 0.0405;
        double pXYY_YYY = 0.3380;
        double pXYY_XYY = 0.41400;
        double pXYY_XXY = 0.20700;
        double pXYY_YYZ = 0.000250;

        for (int k = 0; k < nrOfPatterns; k++) {
            int u = k * nrOfStates;
            partials3[u + sAAA] = pXXX_XXX * ps[u + sAAA] + pXXX_XXY * (ps[u + sAAC] + ps[u + sAAG] + ps[u + sAAT]);
            partials3[u + sCCC] = pXXX_XXX * ps[u + sCCC] + pXXX_XXY * (ps[u + sACC] + ps[u + sCCG] + ps[u + sCCT]);
            partials3[u + sGGG] = pXXX_XXX * ps[u + sGGG] + pXXX_XXY * (ps[u + sAGG] + ps[u + sCGG] + ps[u + sGGT]);
            partials3[u + sTTT] = pXXX_XXX * ps[u + sTTT] + pXXX_XXY * (ps[u + sATT] + ps[u + sCTT] + ps[u + sGTT]);

            partials3[u + sACC] = pXYY_XXX * ps[u + sAAA] + pXYY_YYY * ps[u + sCCC] +
                    pXYY_XYY * ps[u + sACC] + pXYY_XXY * ps[u + sAAC] + pXYY_YYZ * (ps[u + sCCG] + ps[u + sCCT]);
            partials3[u + sAAC] = pXYY_YYY * ps[u + sAAA] + pXYY_XXX * ps[u + sCCC] +
                    pXYY_XXY * ps[u + sACC] + pXYY_XYY * ps[u + sAAC] + pXYY_YYZ * (ps[u + sAAG] + ps[u + sAAT]);

            partials3[u + sAGG] = pXYY_XXX * ps[u + sAAA] + pXYY_YYY * ps[u + sGGG] +
                    pXYY_XYY * ps[u + sAGG] + pXYY_XXY * ps[u + sAAG] + pXYY_YYZ * (ps[u + sCGG] + ps[u + sGGT]);
            partials3[u + sAAG] = pXYY_YYY * ps[u + sAAA] + pXYY_XXX * ps[u + sGGG] +
                    pXYY_XXY * ps[u + sAGG] + pXYY_XYY * ps[u + sAAG] + pXYY_YYZ * (ps[u + sAAC] + ps[u + sAAT]);

            partials3[u + sATT] = pXYY_XXX * ps[u + sAAA] + pXYY_YYY * ps[u + sTTT] +
                    pXYY_XYY * ps[u + sATT] + pXYY_XXY * ps[u + sAAT] + pXYY_YYZ * (ps[u + sCTT] + ps[u + sGTT]);
            partials3[u + sAAT] = pXYY_YYY * ps[u + sAAA] + pXYY_XXX * ps[u + sTTT] +
                    pXYY_XXY * ps[u + sATT] + pXYY_XYY * ps[u + sAAT] + pXYY_YYZ * (ps[u + sAAC] + ps[u + sAAG]);

            partials3[u + sCGG] = pXYY_XXX * ps[u + sCCC] + pXYY_YYY * ps[u + sGGG] +
                    pXYY_XYY * ps[u + sCGG] + pXYY_XXY * ps[u + sCCG] + pXYY_YYZ * (ps[u + sAGG] + ps[u + sGGT]);
            partials3[u + sCCG] = pXYY_YYY * ps[u + sCCC] + pXYY_XXX * ps[u + sGGG] +
                    pXYY_XXY * ps[u + sCGG] + pXYY_XYY * ps[u + sCCG] + pXYY_YYZ * (ps[u + sACC] + ps[u + sCCT]);

            partials3[u + sCTT] = pXYY_XXX * ps[u + sCCC] + pXYY_YYY * ps[u + sTTT] +
                    pXYY_XYY * ps[u + sCTT] + pXYY_XXY * ps[u + sCCT] + pXYY_YYZ * (ps[u + sATT] + ps[u + sGTT]);
            partials3[u + sCCT] = pXYY_YYY * ps[u + sCCC] + pXYY_XXX * ps[u + sTTT] +
                    pXYY_XXY * ps[u + sCTT] + pXYY_XYY * ps[u + sCCT] + pXYY_YYZ * (ps[u + sACC] + ps[u + sCCG]);

            partials3[u + sGTT] = pXYY_XXX * ps[u + sGGG] + pXYY_YYY * ps[u + sTTT] +
                    pXYY_XYY * ps[u + sGTT] + pXYY_XXY * ps[u + sGGT] + pXYY_YYZ * (ps[u + sATT] + ps[u + sCTT]);
            partials3[u + sGGT] = pXYY_YYY * ps[u + sGGG] + pXYY_XXX * ps[u + sTTT] +
                    pXYY_XXY * ps[u + sGTT] + pXYY_XYY * ps[u + sGGT] + pXYY_YYZ * (ps[u + sAGG] + ps[u + sCGG]);
        }
    }



     */