package pomocov.utils;


public class BetaBinomial {

    public static double [] probabilities(int n, double a, double b) {
        double p0 = 1.0;
        double pn = 1.0;
        double [] probs = new double[n+1];
        for (int i = 0; i < n; i++) {
            p0 *= (i + b);
            p0 /= (i + a + b);
            pn *= (i + a);
            pn /= (i + a + b);
        }
        probs[0] = p0;
        probs[n] = pn;
        for (int i = 1; i <= n/2; i++) {
            probs[i] = probs[i-1] * (n + 1 - i) * (i - 1 + a);
            probs[i] /= (n - i + b) * i;
            probs[n-i] = probs[n-i+1] * (n + 1 - i) * (i - 1 + b);
            probs[n-i] /= (n - i + a) * i;
            // if n is even, this calculates probs[n/2] twice
        }
        return probs;
    }

    public static void test() {
        double[] x = probabilities(5, 0.1, 0.3);
        assert Math.abs(x[0] - 0.6330586922268912) < 1e-12;
        assert Math.abs(x[1] - 0.0736114758403361) < 1e-12;
        assert Math.abs(x[2] - 0.0490743172268907) < 1e-12;
        assert Math.abs(x[3] - 0.0448069852941176) < 1e-12;
        assert Math.abs(x[4] - 0.0534237132352941) < 1e-12;
        assert Math.abs(x[5] - 0.1460248161764708) < 1e-12;

        double[] y = probabilities(4, 0.0001, 0.0003);
        assert Math.abs(y[0] - 7.49862548315271e-01) < 1e-12;
        assert Math.abs(y[1] - 9.99716759411087e-05) < 1e-15;
        assert Math.abs(y[2] - 7.49750085802400e-05) < 1e-15;
        assert Math.abs(y[3] - 9.99416939327122e-05) < 1e-15;
        assert Math.abs(y[4] - 2.49862563306118e-01) < 1e-12;

        // test is against R:
        // library(VGAM)
        // dbetabinom.ab(0:5, 5, 0.1, 0.3)
        // dbetabinom.ab(0:4, 4, 0.0001, 0.0003)
    }

}
