
/* GRJ, Apr/May 2020
 */


package pomocov.utils;

import pomocov.AlignmentPoMo;
import pomocov.NodePoMo;
import pomocov.PoMoCoVTreeLikelihood;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Formatter;
import java.util.Locale;

public class DebugTune {






    public static String padStringTo(String s, int n) {
        String ss = new String(s);
        int add = n - ss.length();
        for (int i = 0; i < add; i++) {
            ss += " ";
        }
        return ss;
    }


    public static String nonnegIn8Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        if (x < 0) {
            formatter.format(" neg    ");
        } else if (x <= 1e-10) {
            formatter.format(" 0      ");
        } else if (x < 1e-3) {
            formatter.format("%8.2e", x);
        } else if (x < 9.999) {
            formatter.format("%8.5f", x);
        } else if (x < 99.99) {
            formatter.format("%8.4f", x);
        } else if (x < 999.9) {
            formatter.format("%8.3f", x);
        } else if (x < 9999) {
            formatter.format("%8.2f", x);
        } else {
            formatter.format("%8.0f", x);
        }
        return s.toString();
    }


    public static String floatIn9Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        String sign = " ";
        if (x < 0) {
            sign = "-";
            x = -x;
        }
        if (x == 0) {
            formatter.format("%2s0      ", sign);
        } else if (x <= 1e-99) {
            formatter.format("%2s%7.0e", sign, x);
        } else if (x < 1e-3) {
            formatter.format("%2s%7.1e", sign, x);
        } else if (x < 9.999) {
            formatter.format("%2s%7.5f", sign, x);
        } else if (x < 99.99) {
            formatter.format("%2s%7.4f", sign, x);
        } else if (x < 999.9) {
            formatter.format("%2s%7.3f", sign, x);
        } else if (x < 9999) {
            formatter.format("%2s%7.2f", sign, x);
        } else {
            formatter.format("%2s%7.0f", sign, x);
        }
        return s.toString();
    }



    public static String probabilityIn9Chars(double x) {
        assert x >= 0.0;
        assert x <= 1.0;
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);

        if (x == 0) {
            formatter.format(" 0       ");
        } else if (x <= 1e-99) {
            formatter.format("%9.1e", x);
        } else if (x < 1e-3) {
            formatter.format("%9.2e", x);
        } else {
            formatter.format("%9.6f", x);
        }
        return s.toString();
    }

    public static String floatIn15Chars(double x) {
        StringBuilder s = new StringBuilder();
        Formatter formatter = new Formatter(s, Locale.US);
        formatter.format("%15.8f", x);
        return s.toString();
    }


    public static void printBriefNonNeg1DMatrix(PrintStream ps, String name, double [] M) {
        int n = (int) Math.floor(0.5 + Math.sqrt(M.length));
        assert n*n == M.length;
        ps.println(name);
        for (int row = 0; row < n; row++) {
            for (int col = 0; col < n; col++) {
                ps.print(nonnegIn8Chars(M[n * row + col]) + " ");
            }
            ps.println("");
        }
        ps.println("");
    }

    public static void printNonNeg1DMatrix(PrintStream ps, String name, double [] M) {
        int n = (int) Math.floor(0.5 + Math.sqrt(M.length));
        assert n*n == M.length;
        ps.println(name);
        for (int row = 0; row < n; row++) {
            for (int col = 0; col < n; col++) {
                ps.print(floatIn15Chars(M[n * row + col]) + " ");
            }
            ps.println("");
        }
        ps.println("");
    }


    public static void printSigned2DMatrix(PrintStream ps, String name, double [][] M) {
        int n = M.length;
        assert n > 0;
        assert n == M[0].length;
        ps.println(name);
        for (int row = 0; row < n; row++) {
            for (int col = 0; col < n; col++) {
                ps.print(floatIn9Chars(M[row][col]));
            }
            ps.println("");
        }
        ps.println("");
    }


    private static void printOnePartialOrFrequencies(PrintStream ps, int vPop, double[] pOrF, int offset) {
        int nInEdge = vPop - 1;
        for (int nuc = 0; nuc < 4; nuc++) {
            ps.print(probabilityIn9Chars(pOrF[offset + nuc]));
        }
        ps.print("  ");
        for (int din = 0; din < 6; din++) {
            for (int j = 0; j < nInEdge; j++) {
                ps.print(probabilityIn9Chars(pOrF[offset + 4 + din * nInEdge + j]));
            }
            ps.print("  ");
        }
        ps.println("");
    }


    private static void printPartials(PrintStream ps, int patternCount, int vPop, double[] partials) {
        int nStates = IndicesEtc.nPoMoStates(vPop);
        for (int i = 0; i < patternCount; i++) {
            ps.print("[" + i + "] ");
            printOnePartialOrFrequencies(ps, vPop, partials, i*nStates);
        }
    }


    public static void printNodePartials(PrintStream ps, String eventText, int patternCount, int vPop, double[] partials) {
        ps.println(eventText);
        if (vPop == 3) {
            ps.println("      AAA      CCC      GGG      TTT        ACC       AAC       AGG      AAG       ATT       AAT        CGG      CCG        CTT      CCT       GTT       GGT");
        }
        printPartials(ps, patternCount, vPop, partials);
    }


    public static void printRootFrequencies(PrintStream ps, int index, int vPop, double[] frequencies) {
        ps.println("Root frequencies for index " + index);
        printOnePartialOrFrequencies(ps, vPop, frequencies, 0);
    }


    public static String acgtCountAsText(int [] acgt) {
        assert acgt.length == 4;
        String countText = "";
        if (acgt[0] > 0) {
            countText += ("A" + acgt[0]);
        }
        if (acgt[1] > 0) {
            countText += ("C" + acgt[1]);
        }
        if (acgt[2] > 0) {
            countText += ("G" + acgt[2]);
        }
        if (acgt[3] > 0) {
            countText += ("T" + acgt[3]);
        }
        while (countText.length() < 8) {
            countText += " ";
        }
        countText += " ";
        return countText;
    }


    public static void printSampleAndPartials(PrintStream ps, String eventText, int patternCount, int vPop, ArrayList<int[]> counts, double[] partials) {
        ps.println(eventText);
        if (vPop == 3) {
            ps.println("                AAA      CCC      GGG      TTT        ACC       AAC       AGG      AAG       ATT       AAT        CGG      CCG        CTT      CCT       GTT       GGT");
        }
        int nStates = IndicesEtc.nPoMoStates(vPop);
        for (int i = 0; i < patternCount; i++) {
            ps.print("[" + i + "] " + acgtCountAsText(counts.get(i)));
            printOnePartialOrFrequencies(ps, vPop, partials, i*nStates);
        }
    }



    public static void printPatternWeightsLLhoods(PrintStream ps, int [] weights, double[] patternLLhoods) {
        ps.println("patternWeights  patternLLhoods  product");
        for (int i = 0; i < weights.length; i++) {
            int w = weights[i];
            double lhood = patternLLhoods[i];
            ps.println("[" + i + "] " + w + "  " + lhood + " " + w*lhood);
        }
        ps.println("");
    }
}
